from typing import Iterable, Sequence
from pipeteer import WorkflowQueues
from q.api import ReadQueue

def flatten_queues(queues: WorkflowQueues, prefix: tuple[str, ...] = ()) -> Iterable[tuple[Sequence[str], ReadQueue]]:
  # if 'Qin' in queues and isinstance(queues['Qin'], ReadQueue):
  #   yield prefix + ('Qin',), queues['Qin']
  for id, q in queues.items():
    if isinstance(q, dict):
      yield from flatten_queues(q, prefix + (id,)) # type: ignore
    elif isinstance(q, ReadQueue):
      yield prefix + (id,), q


from .specs import Pipeline, Workflow, Stateful
from .queues import make_queues, MakeQueue, PipelineQueues, WorkflowQueues
from .util import flatten_queues

__all__ = [
  'Pipeline', 'Workflow', 'Stateful',
  'make_queues', 'MakeQueue', 'PipelineQueues', 'WorkflowQueues',
  'flatten_queues',
]


# BEGIN
from typing import Unpack, TypedDict
from q.api import WriteQueue, ReadQueue, Queue
from pipeteer import MakeQueue, make_queues as _make_queues, PipelineQueues
# UNCOMMENT from MODULE import MOD_IMPORTS
INPUTS = str; OUTPUTS = str; INPUT_TYPE = str; OUTPUT_TYPE = str; STATEFUL_INPUT = str # DELETE
class SUBFLOW_CLASS: # DELETE
  Queues = int # DELETE
class STATEFUL_PIPELINE: # DELETE
  Queues = int # DELETE

# LOOP PIPELINE_CLASS INPUTS OUTPUTS
class PIPELINE_CLASS:
  In = INPUTS
  Out = OUTPUTS
  QueueIn = ReadQueue[INPUTS]
  QueueOut = WriteQueue[OUTPUTS]
  Queues = PipelineQueues[INPUTS, OUTPUTS]

# END

# LOOP STATEFUL_CLASS STATEFUL_INPUT STATEFUL_PIPELINE
class STATEFUL_CLASS:
  class Queues(TypedDict):
    Qin: Queue[STATEFUL_INPUT]
    internal: 'STATEFUL_PIPELINE.Queues'

# END

class WORKFLOW:
  class InternalQueues(TypedDict):
    # LOOP SUBFLOW_ID SUBFLOW_CLASS
    SUBFLOW_ID: SUBFLOW_CLASS.Queues
    # END
    # LOOP PIPELINE_ID PIPELINE_CLASS
    PIPELINE_ID: PIPELINE_CLASS.Queues  
    # END
    # LOOP STATEFUL_ID STATEFUL_CLASS
    STATEFUL_ID: STATEFUL_CLASS.Queues
    # END

  class Queues(TypedDict):
    Qin: Queue[INPUT_TYPE]
    internal: 'WORKFLOW.InternalQueues'

  @staticmethod
  def make_queues(make_queue: MakeQueue, output_queue: WriteQueue[OUTPUT_TYPE]) -> Queues:
    VARIABLE = ... # type: ignore # DELETE
    return _make_queues(VARIABLE, make_queue, output_queue) # type: ignore
  
  @staticmethod
  def artifacts(**queues: Unpack['WORKFLOW.InternalQueues']):
    ...
    
# END
from typing import Sequence
from haskellian import Thunk
from templang import parse
from pipeteer import Workflow, Stateful

source = Thunk(lambda: open(__file__).read())

def union_type(types: Sequence[type]):
  return  ' | '.join(t.__name__ for t in types)

def codegen(workflow: Workflow, *, variable: str, module: str, classname: str) -> str:

  translations = {
    'MODULE': module,
    'WORKFLOW': classname,
    'VARIABLE': variable,
    'INPUT_TYPE': workflow.input_type.__name__,
    'OUTPUT_TYPE': union_type(workflow.output_types),
    'SUBFLOW_CLASS': [],
    'SUBFLOW_ID': [],
    'STATEFUL_CLASS': [],
    'STATEFUL_PIPELINE': [],
    'STATEFUL_ID': [],
    'STATEFUL_INPUT': [],
    'PIPELINE_CLASS': [],
    'PIPELINE_ID': [],
    'INPUTS': [],
    'OUTPUTS': [],
  }

  imports = [variable, workflow.input_type.__name__] + [t.__name__ for t in workflow.output_types]

  for id, pipe in workflow.pipelines.items():
    match pipe:
      case Workflow():
        translations['SUBFLOW_CLASS'].append(id.title())
        translations['SUBFLOW_ID'].append(id)
        imports.append(id.title())
      case Stateful():
        translations['STATEFUL_CLASS'].append(id.title() + 'Stateful')
        translations['STATEFUL_ID'].append(id)
        translations['STATEFUL_INPUT'].append(pipe.input_type.__name__)
        translations['STATEFUL_PIPELINE'].append(id.title())
        imports.extend([pipe.input_type.__name__, id.title()])
      case _:
        translations['INPUTS'].append(pipe.input_type.__name__)
        translations['OUTPUTS'].append(union_type(pipe.output_types))
        translations['PIPELINE_CLASS'].append(id.title() + 'Pipeline')
        translations['PIPELINE_ID'].append(id)
        imports.append(pipe.input_type.__name__)
        imports.extend([t.__name__ for t in pipe.output_types])

  translations['MOD_IMPORTS'] = ', '.join(set(imports))

  return parse(source(), translations)
  


from typing import Mapping, Protocol, Generic, TypeVar, Sequence, Union, Callable, Self
import os

A = TypeVar('A', covariant=True)
B = TypeVar('B', covariant=True)
S1 = TypeVar('S1')
S2 = TypeVar('S2')

class Pipeline(Generic[A, B]):
  def __init__(self, input_type: type[A], *output_types: type[B]):
    self.input_type = input_type
    self.output_types: Sequence[type] = output_types

  @property
  def output_type(self) -> type[B]:
    return Union[*self.output_types] # type: ignore

class Stateful(Pipeline[S1, S2], Generic[S1, S2, A, B]):
  def __init__(
    self, input_type: type[S1], output_types: Sequence[type[S2]],
    pipeline: Pipeline[A, B],
    pre: Callable[[S1], A], post: Callable[[S1, B], S2]
  ):
    self.pipeline = pipeline
    self.input_type = input_type
    self.output_types = output_types
    self.pre = pre
    self.post = post

class Workflow(Pipeline[A, B], Generic[A, B]):
  input_task: str
  pipelines: Mapping[str, Pipeline]

  def __init__(self, input_task: str, /, *output_types: type[B], pipelines: Mapping[str, Pipeline]):
    self.input_task = input_task
    self.output_types = output_types
    self.pipelines = pipelines

  @property
  def input_type(self):
    return self.pipelines[self.input_task].input_type
  
  def next_tasks(self, output_types: Sequence[type]) -> Sequence[str]:
    tasks = [
      task for task, pipeline in self.pipelines.items()
        if pipeline.input_type in output_types
    ]
    if set(output_types) & set(self.output_types):
      tasks.append('output')
    return tasks
  
  def codegen(
    self, __file__: str, *,
    variable: str = 'workflow',
    classname: str = 'Workflow',
    overwrite: bool = False
  ):
    folder, file = os.path.split(__file__)
    module, py = os.path.splitext(file)
    outfile = os.path.join(folder, f'{module}_codegen{py}')

    from ._codegen import codegen
    source = codegen(self, variable=variable, module=f'.{module}', classname=classname)

    try:
      with open(outfile, 'w' if overwrite else 'x') as f:
        f.write(source)
    except FileExistsError:
      print('Codegen file already exists. Delete it to regenerate. File:', outfile)


from typing import Mapping, Protocol, TypedDict, Generic, Sequence, TypeVar, TypeAlias, Callable, NamedTuple
from functools import cache
from haskellian import Left, Either, either as E
from q.api import Queue, ReadQueue, WriteQueue, QueueError, ops
from .specs import Pipeline, Workflow, Stateful

A = TypeVar('A')
B = TypeVar('B')
S1 = TypeVar('S1')
S2 = TypeVar('S2')

class PipelineQueues(TypedDict, Generic[A, B]):
  Qin: Queue[A]
  Qout: WriteQueue[B]

class NestedQueues(TypedDict, Generic[A, B]):
  Qin: WriteQueue[A]
  internal: Mapping[str, 'WorkflowQueues']

WorkflowQueues: TypeAlias = PipelineQueues[A, B] | NestedQueues[A, B]

class prejoin(WriteQueue[A], Generic[A]):
  def __init__(self, queues: Sequence[tuple[WriteQueue[A], Sequence[type[A]]]]):
    self.queues = queues
  
  async def push(self, key: str, value: A) -> Either[QueueError, None]:
    for q, types in self.queues:
      for t in types:
        if isinstance(value, t):
          return await q.push(key, value)
    return Left(QueueError(f'Invalid type {type(value).__name__} for {value}'))

class state_merged(WriteQueue[B], Generic[B]):
  def __init__(
    self, post: Callable[[S1, B], S2],
    Qin: ReadQueue[S1], Qout: WriteQueue[S2]
  ):
    self.post = post
    self.Qin = Qin
    self.Qout = Qout

  @E.do[QueueError]()
  async def push(self, key: str, value: B): # type: ignore (python, bruh)
    state = (await self.Qin.read(key)).unsafe()
    next = self.post(state, value)
    (await self.Qout.push(key, next)).unsafe()
    (await self.Qin.pop(key)).unsafe()

class MakeQueue(Protocol):
  def __call__(self, id: Sequence[str], type: type[A], /) -> Queue[A]:
    ...

def make_queues(
  workflow: Workflow[A, B],
  make_queue: MakeQueue,
  output_queue: WriteQueue[B]
) -> WorkflowQueues[A, B]:
  
  def _input_queue(task: Pipeline[A, B], prefix: tuple[str, ...]) -> tuple[Queue[A], Sequence[type[A]]]:
    match task:
      case Workflow():
        return _input_queue(task.pipelines[task.input_task], prefix + (task.input_task,))
      case Pipeline():
        return make_queue(prefix, task.input_type), [task.input_type]
  
  @cache
  def _make_queues(task: Pipeline[A, B], prefix: tuple[str, ...], output_queue: WriteQueue[B]) -> WorkflowQueues[A, B]:
    match task:
      case Workflow():
        def input_of(id: str) -> tuple[WriteQueue, Sequence[type]]:
          if id == 'output':
            return output_queue, task.output_types
          else:
            return _make_queues(task.pipelines[id], prefix + (id,), output_queue)['Qin'], [task.pipelines[id].input_type]

        queues = {
          id: _make_queues(
            pipe, prefix + (id,),
            prejoin([
              input_of(id)
              for id in task.next_tasks(pipe.output_types)
            ])
          )
          for id, pipe in task.pipelines.items()
        }
        return NestedQueues(
          Qin=queues[task.input_task]['Qin'],
          internal=queues
        )
      case Stateful():
        Qstate = make_queue(prefix, task.input_type)
        Qout = state_merged(task.post, Qstate, output_queue)
        queues = dict(_make_queues(task.pipeline, prefix + ('internal',), Qout))
        Qin: Queue = queues['Qin'] # type: ignore
        return NestedQueues(
          Qin=ops.tee(Qstate, Qin.premap(task.pre)),
          internal=queues # type: ignore
        )
      case Pipeline():
        return PipelineQueues(
          Qin=_input_queue(task, prefix)[0],
          Qout=output_queue
        )

  return _make_queues(workflow, (), output_queue)
